from .beta_mode import beta_mode
from .maxent import maxent
from .mle import mle
from .quartile import quartile
from .quartile_int import quartile_int
from .roulette import roulette

__all__ = ["beta_mode", "maxent", "mle", "roulette", "quartile", "quartile_int"]



import logging

from ..distributions import Beta
from ..internal.optimization import optimize_beta_mode

_log = logging.getLogger("preliz")


def beta_mode(lower, upper, mode, mass=0.94, plot=True, plot_kwargs=None, ax=None):
    """Fits Parameters to a Beta Distribution based on the mode, confidence intervals
    and mass of the distribution.

    Parameters
    -----------
    lower : float
        Lower end-point between 0 and upper.
    upper : float
        Upper end-point between lower and 1.
    mode : float
        Mode of the Beta distribution between lower and upper.
    mass : float
        Probability mass between ``lower`` and ``upper`` bounds. Defaults to 0.94
    plot : bool
        Whether to plot the distribution. Defaults to True.
    plot_kwargs : dict
        Dictionary passed to the method ``plot_pdf()``.
    ax : matplotlib axes

    Returns
    --------
    dist : Preliz Beta distribution.
        Beta distribution with fitted parameters alpha and beta for the given mass and intervals.

    See also
    --------
    maxent : maximum entropy distribution with mass in the lower-upper interval.

    References
    ----------
    Adapted from  Evans et al. (2017) see https://doi.org/10.3390/e19100564
    """

    if not 0 < mass <= 1:
        raise ValueError("mass should be larger than 0 and smaller or equal to 1")

    if mode <= lower or mode >= upper:
        raise ValueError("mode should be between lower and upper")

    if upper <= lower:
        raise ValueError("upper should be larger than lower")

    if plot_kwargs is None:
        plot_kwargs = {}

    dist = Beta()
    dist._fit_moments(  # pylint:disable=protected-access
        mean=(upper + lower) / 2, sigma=((upper - lower) / 4) / mass
    )
    tau_a = (dist.alpha - 1) / mode
    tau_b = (dist.beta - 1) / (1 - mode)
    tau = (tau_a + tau_b) / 2
    prob = dist.cdf(upper) - dist.cdf(lower)

    optimize_beta_mode(lower, upper, tau, mode, dist, mass, prob)

    if plot:
        ax = dist.plot_pdf(**plot_kwargs)
        if plot_kwargs.get("pointinterval"):
            cid = -4
        else:
            cid = -1
        ax.plot([lower, upper], [0, 0], "o", color=ax.get_lines()[cid].get_c(), alpha=0.5)

    return ax, dist



import pandas as pd
import numpy as np
import pymc as pm
import bambi as bmb
import preliz as pz


data = pz.Normal(0, 1).rvs(200)

with pm.Model() as model:
    a = pm.Normal("a", mu=0, sigma=1)
    b = pm.HalfNormal("b", sigma=1)
    y = pm.Normal("y", mu=a, sigma=b, observed=data)
    idata = pm.sample(tune=200, draws=500, random_seed=2945)


def test_p2p_pymc():
    pz.posterior_to_prior(model, idata)
    assert 'Gamma\x1b[0m("b", alpha=' in pz.posterior_to_prior(model, idata, alternative="auto")
    pz.posterior_to_prior(model, idata, alternative=[pz.LogNormal()])
    assert 'Gamma\x1b[0m("b", mu=' in pz.posterior_to_prior(
        model, idata, alternative={"b": [pz.Gamma(mu=0)]}
    )


bmb_data = pd.DataFrame(
    {
        "y": np.random.normal(size=117),
        "x": np.random.normal(size=117),
        "x1": np.random.normal(size=117),
    }
)
bmb_prior = {"Intercept": bmb.Prior("HalfStudentT", nu=1)}
bmb_model = bmb.Model("y ~ x + x1", bmb_data, priors=bmb_prior)
bmb_idata = bmb_model.fit(tune=200, draws=200, random_seed=2945)


def test_p2p_bambi():
    pz.posterior_to_prior(bmb_model, bmb_idata)
    assert 'Gamma\x1b[0m", alpha=' in pz.posterior_to_prior(
        bmb_model, bmb_idata, alternative="auto"
    )
    pz.posterior_to_prior(bmb_model, bmb_idata, alternative=[pz.LogNormal()])
    assert 'Normal\x1b[0m", mu=' in pz.posterior_to_prior(
        bmb_model, bmb_idata, alternative={"Intercept": [pz.Normal(mu=1, sigma=1)]}
    )



import numpy as np
import pandas as pd
import preliz as pz
import pymc as pm
import pytest
from numpy.testing import assert_allclose

np.random.seed(42)


@pytest.mark.parametrize(
    "params",
    [
        {
            "mu_x": 0,
            "sigma_x": 10,
            "sigma_z": 10,
            "target": pz.Normal(mu=174, sigma=20),
            "X": np.random.normal(0, 10, 120),
            "new_mu_x": 173.998287,
            "new_sigma_x": 1.946641,
            "new_sigma_z": 19.926345,
        },
        {
            "mu_x": [0, 1],
            "sigma_x": [10, 10],
            "sigma_z": 10,
            "target": pz.Normal(mu=174, sigma=20),
            "X": np.random.normal(0, 10, 120),
            "new_mu_x": (174.089111, 173.908702),
            "new_sigma_x": (2.679288, 2.760424),
            "new_sigma_z": 19.83069,
        },
        {
            "mu_x": 0,
            "sigma_x": 10,
            "sigma_z": 10,
            "target": [(pz.Normal(mu=174, sigma=20), 0.5), (pz.Normal(mu=176, sigma=19.5), 0.5)],
            "X": np.random.normal(0, 10, 120),
            "new_mu_x": 174.852283,
            "new_sigma_x": 1.794118,
            "new_sigma_z": 19.683396,
        },
        {
            "mu_x": [0, 1],
            "sigma_x": [10, 10],
            "sigma_z": 10,
            "target": [
                (pz.Normal(mu=174, sigma=20), 0.5),
                (pz.Normal(mu=176, sigma=19.5), 0.4),
                (pz.StudentT(mu=174, sigma=20, nu=3), 0.1),
            ],
            "X": np.random.normal(0, 10, 120),
            "new_mu_x": (174.75759, 174.775254),
            "new_sigma_x": (2.728134, 2.978235),
            "new_sigma_z": 21.247561,
        },
    ],
)
def test_ppe(params):
    Y = 2 + np.random.normal(params["X"], 1)
    target = params.pop("target")
    with pm.Model() as model:
        x = pm.Normal("x", mu=params["mu_x"], sigma=params["sigma_x"])
        z = pm.HalfNormal("z", params["sigma_z"])
        x_idx = (
            x
            if max(np.asarray(params["mu_x"]).size, np.asarray(params["sigma_x"]).size) == 1
            else x[np.repeat(np.arange(2), Y.size / 2)]
        )
        y = pm.Normal("y", x_idx, z, observed=Y)
    prior, new_prior, pymc_string = pz.ppe(model, target)
    assert_allclose(new_prior["x"].mu, params["new_mu_x"], 1e-6)
    assert_allclose(new_prior["x"].sigma, params["new_sigma_x"], 1e-6)
    assert_allclose(new_prior["z"].sigma, params["new_sigma_z"], 1e-6)


@pytest.mark.parametrize(
    "params",
    [
        {
            "mu_a": 0,
            "sigma_a": 10,
            "sigma_b": 10,
            "sigma_z": 10,
            "target": pz.Normal(mu=40, sigma=7),
            "new_mu_a": 40.00908,
            "new_sigma_a": 0.167701,
            "new_sigma_b": 7.003267,
            "new_sigma_z": 7.003267,
        },
    ],
)
def test_ppe_testdata(params):
    csv_data = pd.read_csv(r"testdata/chemical_shifts_theo_exp.csv")
    diff = csv_data.theo - csv_data.exp
    cat_encode = pd.Categorical(csv_data["aa"])
    idx = cat_encode.codes
    coords = {"aa": cat_encode.categories}
    with pm.Model(coords=coords) as model:
        # hyper_priors
        a = pm.Normal("a", mu=0, sigma=10)
        b = pm.HalfNormal("b", 10)
        z = pm.HalfNormal("z", sigma=10)

        x = pm.Normal("x", mu=a, sigma=b, dims="aa")  # No prior for this!

        y = pm.Normal("y", mu=x[idx], sigma=z, observed=diff)
    prior, new_prior, pymc_string = pz.ppe(model, params["target"])
    assert_allclose(new_prior["a"].mu, params["new_mu_a"], 1e-6)
    assert_allclose(new_prior["a"].sigma, params["new_sigma_a"], 1e-6)
    assert_allclose(new_prior["b"].sigma, params["new_sigma_b"], 1e-6)
    assert_allclose(new_prior["z"].sigma, params["new_sigma_z"], 1e-6)



from math import ceil, floor


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import patches

try:
    import ipywidgets as widgets
except ImportError:
    pass
from ..internal.optimization import fit_to_ecdf, get_distributions
from ..internal.plot_helper import check_inside_notebook, representations
from ..internal.distribution_helper import process_extra
from ..distributions import all_discrete, all_continuous


def roulette(x_min=0, x_max=10, nrows=10, ncols=11, dist_names=None, figsize=None):
    """
    Prior elicitation for 1D distribution using the roulette method.

    Draw 1D distributions using a grid as input.

    Parameters
    ----------
    x_min: Optional[float]
        Minimum value for the domain of the grid and fitted distribution
    x_max: Optional[float]
        Maximum value for the domain of the grid and fitted distribution
    nrows: Optional[int]
        Number of rows for the grid. Defaults to 10.
    ncols: Optional[int]
        Number of columns for the grid. Defaults to 11.
    dist_names: list
        List of distributions names to be used in the elicitation. If None, almost all 1D
        distributions available in PreliZ will be used. Some distributions like Uniform or
        Cauchy are omitted by default.
    figsize: Optional[Tuple[int, int]]
        Figure size. If None it will be defined automatically.

    Returns
    -------
    PreliZ distribution

    References
    ----------
    * Morris D.E. et al. (2014) see https://doi.org/10.1016/j.envsoft.2013.10.010
    * See roulette mode http://optics.eee.nottingham.ac.uk/match/uncertainty.php
    """

    check_inside_notebook(need_widget=True)

    (
        w_x_min,
        w_x_max,
        w_ncols,
        w_nrows,
        w_extra,
        w_repr,
        w_distributions,
        w_checkbox_cont,
        w_checkbox_disc,
        w_checkbox_none,
    ) = get_widgets(
        x_min,
        x_max,
        nrows,
        ncols,
        dist_names,
    )

    output = widgets.Output()

    with output:
        x_min = w_x_min.value
        x_max = w_x_max.value
        nrows = w_nrows.value
        ncols = w_ncols.value

        if figsize is None:
            figsize = (8, 6)

        fig, ax_grid, ax_fit = create_figure(figsize)

        coll = create_grid(x_min, x_max, nrows, ncols, ax=ax_grid)
        grid = Rectangles(fig, coll, nrows, ncols, ax_grid)

        def handle_checkbox_change(_):
            dist_names = handle_checkbox_widget(
                w_distributions.options, w_checkbox_cont, w_checkbox_disc, w_checkbox_none
            )
            w_distributions.value = dist_names

        w_checkbox_none.observe(handle_checkbox_change)
        w_checkbox_cont.observe(handle_checkbox_change)
        w_checkbox_disc.observe(handle_checkbox_change)

        def update_grid_(_):
            update_grid(
                fig.canvas,
                w_x_min.value,
                w_x_max.value,
                w_nrows.value,
                w_ncols.value,
                grid,
                ax_grid,
                ax_fit,
            )

        w_x_min.observe(update_grid_)
        w_x_max.observe(update_grid_)
        w_nrows.observe(update_grid_)
        w_ncols.observe(update_grid_)

        def on_leave_fig_(_):
            on_leave_fig(
                fig.canvas,
                grid,
                w_distributions.value,
                w_repr.value,
                w_x_min.value,
                w_x_max.value,
                w_ncols.value,
                w_extra.value,
                ax_fit,
            )

        w_repr.observe(on_leave_fig_)
        w_distributions.observe(on_leave_fig_)

        def on_value_change(change):
            new_a = change["new"]
            if new_a == w_x_max.value:
                w_x_max.value = new_a + 1

        w_x_min.observe(on_value_change, names="value")

        fig.canvas.mpl_connect(
            "button_release_event",
            lambda event: on_leave_fig(
                fig.canvas,
                grid,
                w_distributions.value,
                w_repr.value,
                w_x_min.value,
                w_x_max.value,
                w_ncols.value,
                w_extra.value,
                ax_fit,
            ),
        )

    controls = widgets.VBox([w_x_min, w_x_max, w_nrows, w_ncols, w_extra])
    control_distribution = widgets.VBox([w_checkbox_cont, w_checkbox_disc, w_checkbox_none])
    display(  # pylint:disable=undefined-variable
        widgets.HBox(
            [
                controls,
                w_repr,
                w_distributions,
                control_distribution,
            ]
        )
    )


def create_figure(figsize):
    """
    Initialize a matplotlib figure with two subplots
    """
    fig, axes = plt.subplots(2, 1, figsize=figsize, constrained_layout=True)
    ax_grid = axes[0]
    ax_fit = axes[1]
    ax_fit.set_yticks([])
    fig.canvas.header_visible = False
    fig.canvas.footer_visible = False
    fig.canvas.toolbar_position = "right"

    return fig, ax_grid, ax_fit


def create_grid(x_min=0, x_max=1, nrows=10, ncols=10, ax=None):
    """
    Create a grid of rectangles
    """
    xx = np.arange(ncols)
    yy = np.arange(nrows)

    if ncols < 11:
        num = ncols
    else:
        num = 11

    ax.set(
        xticks=np.linspace(0, ncols - 1, num=num) + 0.5,
        xticklabels=[f"{i:.1f}" for i in np.linspace(x_min, x_max, num=num)],
    )

    coll = np.zeros((nrows, ncols), dtype=object)
    for idx, xi in enumerate(xx):
        for idy, yi in enumerate(yy):
            sq = patches.Rectangle((xi, yi), 1, 1, fill=True, facecolor="0.8", edgecolor="w")
            ax.add_patch(sq)
            coll[idy, idx] = sq

    ax.set_yticks([])
    ax.relim()
    ax.autoscale_view()
    return coll


class Rectangles:
    """
    Clickable rectangles
    Clicked rectangles are highlighted
    """

    def __init__(self, fig, coll, nrows, ncols, ax):
        self.fig = fig
        self.coll = coll
        self.nrows = nrows
        self.ncols = ncols
        self.ax = ax
        self.weights = {k: 0 for k in range(0, ncols)}
        fig.canvas.mpl_connect("button_press_event", self)

    def __call__(self, event):
        if event.inaxes == self.ax:
            x = event.xdata
            y = event.ydata
            idx = floor(x)
            idy = ceil(y)

            if 0 <= idx < self.ncols and 0 <= idy <= self.nrows:
                if self.weights[idx] >= idy:
                    idy -= 1
                    for row in range(self.nrows):
                        self.coll[row, idx].set_facecolor("0.8")
                self.weights[idx] = idy
                for row in range(idy):
                    self.coll[row, idx].set_facecolor("C1")
                self.fig.canvas.draw()


def on_leave_fig(canvas, grid, dist_names, kind_plot, x_min, x_max, ncols, extra, ax):
    x_min = float(x_min)
    x_max = float(x_max)
    ncols = float(ncols)
    x_range = x_max - x_min
    extra_pros = process_extra(extra)

    x_vals, ecdf, mean, std, filled_columns = weights_to_ecdf(grid.weights, x_min, x_range, ncols)

    fitted_dist = None
    if filled_columns > 1:
        selected_distributions = get_distributions(dist_names)

        if selected_distributions:
            reset_dist_panel(x_min, x_max, ax, yticks=False)
            fitted_dist = fit_to_ecdf(
                selected_distributions,
                x_vals,
                ecdf,
                mean,
                std,
                x_min,
                x_max,
                extra_pros,
            )

            if fitted_dist is None:
                ax.set_title("domain error")
            else:
                representations(fitted_dist, kind_plot, ax)
    else:
        reset_dist_panel(x_min, x_max, ax, yticks=True)
    canvas.draw()

    return fitted_dist


def weights_to_ecdf(weights, x_min, x_range, ncols):
    """
    Turn the weights (chips) into the empirical cdf
    """
    step = x_range / (ncols - 1)
    x_vals = [(k + 0.5) * step + x_min for k, v in weights.items() if v != 0]
    total = sum(weights.values())
    probabilities = [v / total for v in weights.values() if v != 0]
    cum_sum = np.cumsum(probabilities)

    mean = sum(value * prob for value, prob in zip(x_vals, probabilities))
    std = (sum(prob * (value - mean) ** 2 for value, prob in zip(x_vals, probabilities))) ** 0.5

    return x_vals, cum_sum, mean, std, len(x_vals)


def update_grid(canvas, x_min, x_max, nrows, ncols, grid, ax_grid, ax_fit):
    """
    Update the grid subplot
    """
    ax_grid.cla()
    coll = create_grid(x_min=x_min, x_max=x_max, nrows=nrows, ncols=ncols, ax=ax_grid)
    grid.coll = coll
    grid.ncols = ncols
    grid.nrows = nrows
    grid.weights = {k: 0 for k in range(0, ncols)}
    reset_dist_panel(x_min, x_max, ax_fit, yticks=True)
    ax_grid.set_yticks([])
    ax_grid.relim()
    ax_grid.autoscale_view()
    canvas.draw()


def reset_dist_panel(x_min, x_max, ax, yticks):
    """
    Clean the distribution subplot
    """
    ax.cla()
    if yticks:
        ax.set_yticks([])
    ax.set_xlim(x_min, x_max)
    ax.relim()
    ax.autoscale_view()


def handle_checkbox_widget(options, w_checkbox_cont, w_checkbox_disc, w_checkbox_none):
    if w_checkbox_none.value:
        w_checkbox_disc.value = False
        w_checkbox_cont.value = False
        return []
    all_cls = []
    if w_checkbox_cont.value:
        all_cont_str = [  # pylint:disable=unnecessary-comprehension
            dist for dist in (cls.__name__ for cls in all_continuous if cls.__name__ in options)
        ]
        all_cls += all_cont_str
    if w_checkbox_disc.value:
        all_dist_str = [  # pylint:disable=unnecessary-comprehension
            dist for dist in (cls.__name__ for cls in all_discrete if cls.__name__ in options)
        ]
        all_cls += all_dist_str
    return all_cls


def get_widgets(x_min, x_max, nrows, ncols, dist_names):

    width_entry_text = widgets.Layout(width="150px")
    width_repr_text = widgets.Layout(width="250px")
    width_distribution_text = widgets.Layout(width="150px", height="125px")

    w_x_min = widgets.FloatText(
        value=x_min,
        step=1,
        description="x_min:",
        disabled=False,
        layout=width_entry_text,
    )

    w_x_max = widgets.FloatText(
        value=x_max,
        step=1,
        description="x_max:",
        disabled=False,
        layout=width_entry_text,
    )

    w_nrows = widgets.BoundedIntText(
        value=nrows,
        min=2,
        step=1,
        description="n_rows:",
        disabled=False,
        layout=width_entry_text,
    )

    w_ncols = widgets.BoundedIntText(
        value=ncols,
        min=2,
        step=1,
        description="n_cols:",
        disabled=False,
        layout=width_entry_text,
    )

    w_extra = widgets.Textarea(
        value="",
        placeholder="Pass extra parameters",
        description="params:",
        disabled=False,
        layout=width_repr_text,
    )

    w_repr = widgets.RadioButtons(
        options=["pdf", "cdf", "ppf"],
        value="pdf",
        description="",
        disabled=False,
        layout=width_entry_text,
    )

    if dist_names is None:

        default_dist = ["Normal", "BetaScaled", "Gamma", "LogNormal", "StudentT"]

        dist_names = [
            "AsymmetricLaplace",
            "BetaScaled",
            "ChiSquared",
            "ExGaussian",
            "Exponential",
            "Gamma",
            "Gumbel",
            "HalfNormal",
            "HalfStudentT",
            "InverseGamma",
            "Laplace",
            "LogNormal",
            "Logistic",
            # "LogitNormal", # fails if we add chips at x_value= 1
            "Moyal",
            "Normal",
            "Pareto",
            "Rice",
            "SkewNormal",
            "StudentT",
            "Triangular",
            "VonMises",
            "Wald",
            "Weibull",
            "BetaBinomial",
            "DiscreteWeibull",
            "Geometric",
            "NegativeBinomial",
            "Poisson",
        ]

    else:
        default_dist = dist_names

    w_distributions = widgets.SelectMultiple(
        options=dist_names,
        value=default_dist,
        description="",
        disabled=False,
        layout=width_distribution_text,
    )

    w_checkbox_cont = widgets.Checkbox(
        value=False, description="Continuous", disabled=False, indent=False
    )
    w_checkbox_disc = widgets.Checkbox(
        value=False, description="Discrete", disabled=False, indent=False
    )
    w_checkbox_none = widgets.Checkbox(
        value=False, description="None", disabled=False, indent=False
    )

    return (
        w_x_min,
        w_x_max,
        w_ncols,
        w_nrows,
        w_extra,
        w_repr,
        w_distributions,
        w_checkbox_cont,
        w_checkbox_disc,
        w_checkbox_none,
    )



import logging

from contextlib import contextmanager


@contextmanager
def disable_pymc_sampling_logs(logger: logging.Logger = logging.getLogger("pymc")):
    effective_level = logger.getEffectiveLevel()
    logger.setLevel(logging.ERROR)
    try:
        yield
    finally:
        logger.setLevel(effective_level)



try:
    import ipywidgets as widgets
except ImportError:
    pass

from preliz.internal.optimization import fit_to_quartile
from ..internal.plot_helper import (
    create_figure,
    check_inside_notebook,
    representations,
    reset_dist_panel,
)
from ..internal.distribution_helper import process_extra


def quartile_int(q1=1, q2=2, q3=3, dist_names=None, figsize=None):
    """
    Prior elicitation for 1D distributions from quartiles.

    Parameters
    ----------
    q1 : float
        First quartile, i.e 0.25 of the mass is below this point.
    q2 : float
        Second quartile, i.e 0.50 of the mass is below this point. This is also know
        as the median.
    q3 : float
        Third quartile, i.e 0.75 of the mass is below this point.
    dist_names: list
        List of distributions names to be used in the elicitation. If None, almost all 1D
        distributions available in PreliZ will be used. Some distributions like Uniform or
        Cauchy are omitted by default.
    figsize: Optional[Tuple[int, int]]
        Figure size. If None it will be defined automatically.

    Note
    ----
    Use the `params` text box to parametrize distributions, for instance write
    `BetaScaled(lower=-1, upper=10)` to specify the upper and lower bounds of BetaScaled
    distribution. To parametrize more that one distribution use commas for example
    `StudentT(nu=3), TruncatedNormal(lower=-2, upper=inf)`

    References
    ----------
    * Morris D.E. et al. (2014) see https://doi.org/10.1016/j.envsoft.2013.10.010
    * See quartile mode http://optics.eee.nottingham.ac.uk/match/uncertainty.php
    """

    check_inside_notebook(need_widget=True)

    w_q1, w_q2, w_q3, w_extra, w_repr, w_distributions = get_widgets(q1, q2, q3, dist_names)

    output = widgets.Output()

    with output:

        if figsize is None:
            figsize = (8, 6)

        fig, ax_fit = create_figure(figsize)

        def match_distribution_(_):
            match_distribution(
                fig.canvas,
                w_distributions.value,
                w_repr.value,
                w_q1.value,
                w_q2.value,
                w_q3.value,
                w_extra.value,
                ax_fit,
            )

        w_repr.observe(match_distribution_)
        w_distributions.observe(match_distribution_)
        w_q1.observe(match_distribution_)
        w_q2.observe(match_distribution_)
        w_q3.observe(match_distribution_)

        fig.canvas.mpl_connect(
            "button_release_event",
            lambda event: match_distribution(
                fig.canvas,
                w_distributions.value,
                w_repr.value,
                w_q1.value,
                w_q2.value,
                w_q3.value,
                w_extra.value,
                ax_fit,
            ),
        )

    controls = widgets.VBox([w_q1, w_q2, w_q3, w_extra])

    display(widgets.HBox([controls, w_repr, w_distributions]))  # pylint:disable=undefined-variable


def match_distribution(canvas, dist_names, kind_plot, q1, q2, q3, extra, ax):
    q1 = float(q1)
    q2 = float(q2)
    q3 = float(q3)
    extra_pros = process_extra(extra)
    fitted_dist = None

    if q1 < q2 < q3:
        reset_dist_panel(ax, yticks=False)

        fitted_dist = fit_to_quartile(dist_names, q1, q2, q3, extra_pros)

        if fitted_dist is None:
            ax.set_title("domain error")
        else:
            representations(fitted_dist, kind_plot, ax)
    else:
        reset_dist_panel(ax, yticks=True)
        ax.set_title("quantiles must follow the order: q1 < q2 < q3 ")

    canvas.draw()

    return fitted_dist


def get_widgets(q1, q2, q3, dist_names=None):

    width_entry_text = widgets.Layout(width="150px")
    width_repr_text = widgets.Layout(width="250px")
    width_distribution_text = widgets.Layout(width="150px", height="125px")

    w_q1 = widgets.FloatText(
        value=q1,
        step=0.1,
        description="q1",
        disabled=False,
        layout=width_entry_text,
    )

    w_q2 = widgets.FloatText(
        value=q2,
        step=0.1,
        description="q2",
        disabled=False,
        layout=width_entry_text,
    )

    w_q3 = widgets.FloatText(
        value=q3,
        step=0.1,
        description="q3",
        disabled=False,
        layout=width_entry_text,
    )

    w_extra = widgets.Textarea(
        value="",
        placeholder="Pass extra parameters",
        description="params:",
        disabled=False,
        layout=width_repr_text,
    )

    w_repr = widgets.RadioButtons(
        options=["pdf", "cdf", "ppf"],
        value="pdf",
        description="",
        disabled=False,
        layout=width_repr_text,
    )

    if dist_names is None:

        default_dist = ["Normal", "BetaScaled", "Gamma", "LogNormal", "StudentT"]

        dist_names = [
            "AsymmetricLaplace",
            "BetaScaled",
            "ChiSquared",
            "ExGaussian",
            "Exponential",
            "Gamma",
            "Gumbel",
            "HalfNormal",
            "HalfStudentT",
            "InverseGamma",
            "Laplace",
            "LogNormal",
            "Logistic",
            "LogitNormal",
            "Moyal",
            "Normal",
            "Pareto",
            "Rice",
            "SkewNormal",
            "StudentT",
            "Triangular",
            "VonMises",
            "Wald",
            "Weibull",
            "BetaBinomial",
            "DiscreteWeibull",
            "Geometric",
            "NegativeBinomial",
            "Poisson",
        ]
    else:
        default_dist = dist_names

    w_distributions = widgets.SelectMultiple(
        options=dist_names,
        value=default_dist,
        description="",
        disabled=False,
        layout=width_distribution_text,
    )

    return w_q1, w_q2, w_q3, w_extra, w_repr, w_distributions



"""Methods to communicate with PyMC."""

# pylint: disable=protected-access
from sys import modules

import numpy as np

try:
    from pytensor.tensor import vector, TensorConstant
    from pytensor.graph.basic import ancestors
    from pymc import logp, compile_pymc
    from pymc.util import is_transformed_name, get_untransformed_name
except ModuleNotFoundError:
    pass

from preliz.internal.optimization import get_distributions


def backfitting(prior, p_model, var_info2):
    """
    Fit the samples from prior into user provided model's prior.
    from the perspective of ppe "prior" is actually an approximated posterior
    but from the users perspective is its prior.
    We need to "backfitted" because we can not use arbitrary samples as priors.
    We need probability distributions.
    """
    new_priors = {}
    for key, size_inf in var_info2.items():
        if not size_inf[2]:
            size = size_inf[1]
            if size > 1:
                params = []
                for i in range(size):
                    value = prior[f"{key}__{i}"]
                    dist = p_model[key]
                    dist._fit_mle(value)
                    params.append(dist.params)
                dist._parametrization(*[np.array(x) for x in zip(*params)])
            else:
                value = prior[key]
                dist = p_model[key]
                dist._fit_mle(value)

            new_priors[key] = dist

    return new_priors


def compile_logp(model):
    """
    Compile the log-likelihood function for the model.
    We need to be able to condition it on parameters or data.
    Because during the optimization routine we need to change both.
    Currently this will fail for a prior that depends on other prior.
    """
    value = vector("value")
    rv_logp = logp(*model.observed_RVs, value)
    rv_logp_fn = compile_pymc([*model.free_RVs, value], rv_logp, on_unused_input="ignore")
    rv_logp_fn.trust_input = True

    def fmodel(params, obs, var_info, p_model):
        params = reshape_params(model, var_info, p_model, params)
        return -rv_logp_fn(*params, obs).sum()

    return fmodel


def get_pymc_to_preliz():
    """
    Generate dictionary mapping pymc to preliz distributions
    """
    all_distributions = [
        dist
        for dist in modules["preliz.distributions"].__all__
        if dist not in ["Truncated", "Censored", "Hurdle"]
    ]
    pymc_to_preliz = dict(
        zip([dist.lower() for dist in all_distributions], get_distributions(all_distributions))
    )
    return pymc_to_preliz


def get_guess(model, free_rvs):
    """
    Get initial guess for optimization routine.
    """
    init = []

    free_rvs_names = [rv.name for rv in free_rvs]
    for key, value in model.initial_point().items():

        if is_transformed_name(key):
            name = get_untransformed_name(key)
            value = model.rvs_to_transforms[model.named_vars[name]].backward(value).eval()
        else:
            name = key

        if name in free_rvs_names:
            init.append(value)

    return np.concatenate([np.atleast_1d(arr) for arr in init]).flatten()


def get_model_information(model):  # pylint: disable=too-many-locals
    """
    Get information from the PyMC model.

    This needs some love. We even have a variable named var_info,
    and another one var_info2!
    """

    bounds = []
    prior = {}
    p_model = {}
    var_info = {}
    var_info2 = {}
    free_rvs = []
    pymc_to_preliz = get_pymc_to_preliz()
    rvs_to_values = model.rvs_to_values

    for r_v in model.free_RVs:
        r_v_eval = r_v.eval()
        size = r_v_eval.size
        shape = r_v_eval.shape
        nc_parents = non_constant_parents(r_v, model.free_RVs)
        name = (
            r_v.owner.op.name if r_v.owner.op.name else str(r_v.owner.op).split("RV", 1)[0].lower()
        )
        dist = pymc_to_preliz[name]
        p_model[r_v.name] = dist
        if nc_parents:
            idxs = [free_rvs.index(var_) for var_ in nc_parents]
            # the keys are the name of the (transformed) variable
            var_info[rvs_to_values[r_v].name] = (shape, size, idxs)
            # the keys are the name of the (untransformed) variable
            var_info2[r_v.name] = (shape, size, idxs)
        else:
            free_rvs.append(r_v)

            if size > 1:
                for i in range(size):
                    bounds.append(dist.support)
                    prior[f"{r_v.name}__{i}"] = []
            else:
                bounds.append(dist.support)
                prior[r_v.name] = []

            # the keys are the name of the (transformed) variable
            var_info[rvs_to_values[r_v].name] = (shape, size, nc_parents)
            # the keys are the name of the (untransformed) variable
            var_info2[r_v.name] = (shape, size, nc_parents)

    draws = model.observed_RVs[0].eval().size

    return bounds, prior, p_model, var_info, var_info2, draws, free_rvs


def write_pymc_string(new_priors, var_info):
    """
    Return a string with the new priors for the PyMC model.
    So the user can copy and paste, ideally with none to minimal changes.
    """

    header = "with pm.Model() as model:\n"

    for key, value in new_priors.items():
        dist_name, dist_params = repr(value).split("(")
        size = var_info[key][1]
        if size > 1:
            dist_params = dist_params.split(")")[0]
            header += f'{key:>4} = pm.{dist_name}("{key}", {dist_params}, shape={size})\n'
        else:
            header += f'{key:>4} = pm.{dist_name}("{key}", {dist_params}\n'

    return header


def reshape_params(model, var_info, p_model, params):
    """
    We flatten the parameters to be able to use them in the optimization routine.
    """
    size = 0
    value = []
    for var in model.value_vars:
        shape, new_size, idxs = var_info[var.name]
        if idxs:
            dist = p_model[var.name]
            dist._parametrization(*params[idxs])
            if new_size > 1:
                value.append(np.repeat(dist.mean(), new_size))
            else:
                value.append(dist.mean())
            size += new_size
        else:
            var_samples = params[size : size + new_size]
            value.append(var_samples.reshape(shape))
            size += new_size

    return value


def non_constant_parents(var_, free_rvs):
    """Find the parents of a variable that are not constant."""
    parents = []
    for variable in var_.get_parents()[0].inputs[2:]:
        if not isinstance(variable, TensorConstant):
            for free_rv in free_rvs:
                if free_rv in list(ancestors([variable])) and free_rv not in parents:
                    parents.append(free_rv)
    return parents



"""
PreliZ.

Tools to help you pick a prior
"""
import logging
from os import path as os_path

from matplotlib import rcParams
from matplotlib import style

from .distributions import *
from .predictive import *
from .ppls import *
from .unidimensional import *
from .multidimensional import *


__version__ = "0.9.1"

_log = logging.getLogger("preliz")

if not logging.root.handlers:
    _log.setLevel(logging.INFO)
    if len(_log.handlers) == 0:
        handler = logging.StreamHandler()
        _log.addHandler(handler)

# Allow legend outside plot in maxent to be included when saving a figure
# We may want to make this more explicit by having preliz.rcParams
rcParams["savefig.bbox"] = "tight"


# add PreliZ's styles to matplotlib's styles
_preliz_style_path = os_path.join(os_path.dirname(__file__), "styles")
style.core.USER_LIBRARY_PATHS.append(_preliz_style_path)
style.core.reload_library()

# clean namespace
del logging, os_path, rcParams, _preliz_style_path, _log



"""Functions to communicate with PPLs."""

import logging

from preliz.internal.parser import get_engine
from preliz.distributions import Gamma, Normal, HalfNormal
from preliz.unidimensional.mle import mle
from preliz.ppls.pymc_io import get_model_information, write_pymc_string
from preliz.ppls.bambi_io import get_bmb_model_information, write_bambi_string

_log = logging.getLogger("preliz")


def posterior_to_prior(model, idata, alternative=None, engine="auto"):
    """
    Fit a posterior from a model to its prior

    The fit is based on maximum likelihood of each posterior marginal to the prior
    in the model. It is expected that the posterior was computed from the model.

    Parameters
    ----------
    model : A PyMC or a Bambi Model
    idata : InferenceData
        InferenceData with a posterior group.
    alternative : "auto", list or dict
        Defaults to None, the samples are fit to the original prior distribution.
        If "auto", the method evaluates the fit to the original prior plus a set of
        predefined distributions.
        Use a list of PreliZ distribution to specify the alternative distributions
        you want to consider.
        Use a dict with variables names in ``model`` as keys and a list of PreliZ
        distributions as values. This allows to specify alternative distributions
        per variable.
    engine : "auto", "pymc" or "bambi"
        Library used to define the model. Either `pymc` or `bambi`. Default is `auto`.
        The function will automatically select the appropriate library to use based on the model
        provided.
    """
    _log.info(""""This is an experimental method under development, use with caution.""")
    engine = get_engine(model) if engine == "auto" else engine
    if engine == "bambi":
        _, _, model_info, _, var_info2, *_ = get_bmb_model_information(model)
    elif engine == "pymc":
        _, _, model_info, _, var_info2, *_ = get_model_information(model)
    new_priors = {}
    posterior = idata.posterior.stack(sample=("chain", "draw"))

    if alternative is None:
        for var, dist in model_info.items():
            dist._fit_mle(posterior[var].values)
            new_priors[var] = dist
    else:
        for var, dist in model_info.items():
            dists = [dist]

            if alternative == "auto":
                dists += [Normal(), HalfNormal(), Gamma()]
            elif isinstance(alternative, list):
                dists += alternative
            elif isinstance(alternative, dict):
                dists += alternative.get(var, [])

            idx, _ = mle(dists, posterior[var].values, plot=False)
            new_priors[var] = dists[idx[0]]
    if engine == "bambi":
        new_model = write_bambi_string(new_priors, var_info2)
    elif engine == "pymc":
        new_model = write_pymc_string(new_priors, var_info2)

    return new_model



from preliz.ppls.pymc_io import get_model_information


def get_bmb_model_information(model):
    if not model.built:
        model.build()
    pymc_model = model.backend.model
    return get_model_information(pymc_model)


def write_bambi_string(new_priors, var_info):
    """
    Return a string with the new priors for the Bambi model.
    So the user can copy and paste, ideally with none to minimal changes.
    """
    header = "{"
    for key, value in new_priors.items():
        dist_name, dist_params = repr(value).split("(")
        dist_params = dist_params.rstrip(")")
        size = var_info[key][1]
        if size > 1:
            header += f'"{key}" : bmb.Prior("{dist_name}", {dist_params}, shape={size}), '
        else:
            header += f'"{key}" : bmb.Prior("{dist_name}", {dist_params}), '

    header = header.rstrip(", ") + "}"
    return header



"""
Optimization routines and utilities
"""
from sys import modules
import warnings
from copy import copy

import numpy as np
from scipy.optimize import minimize, least_squares, root_scalar, brentq
from scipy.special import i0, i1, i0e, i1e  # pylint: disable=no-name-in-module
from .distribution_helper import init_vals as default_vals


def optimize_max_ent(dist, lower, upper, mass, none_idx, fixed):
    def prob_bound(params, dist, lower, upper, mass):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        if dist.kind == "discrete":
            lower -= 1
        cdf0 = dist.cdf(lower)
        cdf1 = dist.cdf(upper)
        loss = (cdf1 - cdf0) - mass
        return loss

    def entropy_loss(params, dist):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        return -dist.entropy()

    cons = {
        "type": "eq",
        "fun": prob_bound,
        "args": (dist, lower, upper, mass),
    }
    init_vals = np.array(dist.params)[none_idx]
    bounds = np.array(dist.params_support)[none_idx]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Values in x were outside bounds")
        opt = minimize(entropy_loss, x0=init_vals, bounds=bounds, args=(dist), constraints=cons)

    params = get_params(dist, opt["x"], none_idx, fixed)
    dist._parametrization(**params)

    return opt


def get_params(dist, params, none_idx, fixed):
    params_ = {}
    pdx = 0
    fdx = 0
    for idx in range(len(dist.params)):
        name = dist.param_names[idx]
        if idx in none_idx:
            params_[name] = params[pdx]
            pdx += 1
        else:
            params_[name] = fixed[fdx]
            fdx += 1

    return params_


def optimize_quartile(dist, x_vals, none_idx, fixed):
    def func(params, dist, x_vals):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        loss = dist.cdf(x_vals) - [0.25, 0.5, 0.75]
        return loss

    init_vals = np.array(dist.params)[none_idx]
    bounds = np.array(dist.params_support)[none_idx]
    bounds = list(zip(*bounds))

    opt = least_squares(func, x0=init_vals, args=(dist, x_vals), bounds=bounds)
    params = get_params(dist, opt["x"], none_idx, fixed)
    dist._parametrization(**params)
    return opt


def optimize_cdf(dist, x_vals, ecdf, none_idx, fixed):
    def func(params, dist, x_vals, ecdf):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        loss = dist.cdf(x_vals) - ecdf
        return loss

    init_vals = np.array(dist.params)[none_idx]
    bounds = np.array(dist.params_support)[none_idx]
    bounds = list(zip(*bounds))

    opt = least_squares(func, x0=init_vals, args=(dist, x_vals, ecdf), bounds=bounds)
    params = get_params(dist, opt["x"], none_idx, fixed)
    dist._parametrization(**params)
    loss = opt["cost"]
    return loss


def optimize_moments(dist, mean, sigma, params=None):
    def func(params, dist, mean, sigma):
        params = get_params(dist, params, none_idx, fixed)
        dist._parametrization(**params)
        loss = abs(dist.mean() - mean) + abs(dist.std() - sigma)
        return loss

    none_idx, fixed = get_fixed_params(dist)

    if params is not None:
        dist._update(*params)
    else:
        name = dist.__class__.__name__
        if name == "Truncated":
            vals = copy(default_vals["Truncated"])
            vals.update(default_vals[dist.dist.__class__.__name__])
            dist._parametrization(**vals)
        else:
            dist._update(**default_vals[name])

    init_vals = np.array(dist.params)[none_idx]

    if dist.__class__.__name__ in ["HyperGeometric", "BetaBinomial"]:
        opt = least_squares(func, x0=init_vals, args=(dist, mean, sigma))
    else:
        bounds = np.array(dist.params_support)[none_idx]
        bounds = list(zip(*bounds))
        if dist.__class__.__name__ in ["DiscreteWeibull"]:
            opt = least_squares(
                func, x0=init_vals, args=(dist, mean, sigma), bounds=bounds, loss="soft_l1"
            )
        else:
            opt = least_squares(func, x0=init_vals, args=(dist, mean, sigma), bounds=bounds)

    params = get_params(dist, opt["x"], none_idx, fixed)
    dist._parametrization(**params)
    return opt


def optimize_moments_rice(mean, std_dev):
    """
    Moment matching for the Rice distribution

    This function uses the Koay inversion technique, see: https://doi.org/10.1016/j.jmr.2006.01.016
    and https://en.wikipedia.org/wiki/Rice_distribution
    """

    ratio = mean / std_dev

    if ratio < 1.913:  # Rayleigh distribution
        nu = np.finfo(float).eps
        sigma = 0.655 * std_dev
    else:

        def xi(theta):
            return (
                2
                + theta**2
                - np.pi
                / 8
                * np.exp(-(theta**2) / 2)
                * ((2 + theta**2) * i0(theta**2 / 4) + theta**2 * i1(theta**2 / 4)) ** 2
            )

        def fpf(theta):
            return (xi(theta) * (1 + ratio**2) - 2) ** 0.5

        def func(theta):
            return np.abs(fpf(theta) - theta)

        theta = minimize(func, x0=fpf(1), bounds=[(0, None)]).x
        xi_theta = xi(theta)
        sigma = std_dev / xi_theta**0.5
        nu = (mean**2 + (xi_theta - 2) * sigma**2) ** 0.5

    return nu, sigma


def optimize_ml(dist, sample):
    def negll(params, dist, sample):
        dist._update(*params)
        return dist._neg_logpdf(sample)

    dist._fit_moments(np.mean(sample), np.std(sample))
    init_vals = dist.params

    opt = minimize(negll, x0=init_vals, bounds=dist.params_support, args=(dist, sample))

    dist._update(*opt["x"])

    return opt


def optimize_dirichlet_mode(lower_bounds, mode, target_mass, _dist):
    def prob_approx(tau, lower_bounds, mode, _dist):
        alpha = [1 + tau * mode_i for mode_i in mode]
        a_0 = sum(alpha)
        marginal_prob_list = []
        for a_i, lbi in zip(alpha, lower_bounds):
            _dist._parametrization(a_i, a_0 - a_i)
            marginal_prob_list.append(_dist.cdf(lbi))

        mean_cdf = np.mean(marginal_prob_list)
        return mean_cdf, alpha

    tau = 1
    new_prob, alpha = prob_approx(tau, lower_bounds, mode, _dist)

    while abs(new_prob - target_mass) > 0.0001:
        if new_prob < target_mass:
            tau -= 0.5 * tau
        else:
            tau += 0.5 * tau

        new_prob, alpha = prob_approx(tau, lower_bounds, mode, _dist)

    return new_prob, alpha


def optimize_beta_mode(lower, upper, tau_not, mode, dist, mass, prob):
    while abs(prob - mass) > 0.0001:
        alpha = 1 + mode * tau_not
        beta = 1 + (1 - mode) * tau_not
        dist._parametrization(alpha, beta)
        prob = dist.cdf(upper) - dist.cdf(lower)

        if prob > mass:
            tau_not -= 0.5 * tau_not
        else:
            tau_not += 0.5 * tau_not


def optimize_hdi(dist, mass):
    def interval_loss(params):
        cdf = dist.cdf(params)
        loss = (cdf[1] - cdf[0]) - mass
        return loss

    def interval_short(params):
        lower, upper = params
        return upper - lower

    cons = {
        "type": "eq",
        "fun": interval_loss,
    }
    init_vals = dist.eti(mass=mass)
    bounds = np.array([dist.support])
    opt = minimize(interval_short, x0=init_vals, bounds=bounds, constraints=cons)

    lower, upper = opt.x
    if dist.kind == "discrete":
        upper = np.floor(upper - 1).astype(int)
        lower = np.ceil(lower).astype(int)

    return lower, upper


def optimize_pymc_model(
    fmodel, target, draws, prior, initial_guess, bounds, var_info, p_model, rng
):
    for _ in range(400):
        # can we sample systematically from these and less random?
        # This should be more flexible and allow other targets than just
        # a preliz distribution
        if isinstance(target, list):
            obs = get_weighted_rvs(target, draws, rng)
        else:
            obs = target.rvs(draws, random_state=rng)
        result = minimize(
            fmodel,
            initial_guess,
            tol=0.001,
            method="SLSQP",
            args=(obs, var_info, p_model),
            bounds=bounds,
        )

        optimal_params = result.x
        initial_guess = optimal_params

        for key, param in zip(prior.keys(), optimal_params):
            prior[key].append(param)

    # convert to numpy arrays
    for key, value in prior.items():
        prior[key] = np.array(value)

    return prior


def relative_error(dist, lower, upper, required_mass):
    if dist.kind == "discrete":
        lower -= 1
    computed_mass = dist.cdf(upper) - dist.cdf(lower)
    return abs((computed_mass - required_mass) / required_mass * 100), computed_mass


def get_distributions(dist_names):
    """
    Generate a subset of distributions which names agree with those in dist_names
    """
    dists = []
    for name in dist_names:
        dist = getattr(modules["preliz"], name)
        dists.append(dist())

    return dists


def fit_to_ecdf(selected_distributions, x_vals, ecdf, mean, std, x_min, x_max, extra_pros):
    """
    Minimize the difference between the cdf and the ecdf over a grid of values
    defined by x_min and x_max

    Note: This function is intended to be used with pz.roulette
    """
    fitted = Loss(len(selected_distributions))
    for dist in selected_distributions:
        if dist.__class__.__name__ in extra_pros:
            dist._parametrization(**extra_pros[dist.__class__.__name__])
        if dist.__class__.__name__ == "BetaScaled":
            update_bounds_beta_scaled(dist, x_min, x_max)

        if dist._check_endpoints(x_min, x_max, raise_error=False):
            none_idx, fixed = get_fixed_params(dist)
            dist._fit_moments(mean, std)  # pylint:disable=protected-access
            loss = optimize_cdf(dist, x_vals, ecdf, none_idx, fixed)

            fitted.update(loss, dist)

    return fitted.dist


def fit_to_sample(selected_distributions, sample, x_min, x_max):
    """
    Maximize the likelihood given a sample
    """
    fitted = Loss(len(selected_distributions))
    sample_size = len(sample)
    for dist in selected_distributions:
        if dist.__class__.__name__ in ["BetaScaled", "TruncatedNormal"]:
            update_bounds_beta_scaled(dist, x_min, x_max)

        loss = np.inf
        if dist._check_endpoints(x_min, x_max, raise_error=False):
            dist._fit_mle(sample)  # pylint:disable=protected-access
            corr = get_penalization(sample_size, dist)
            loss = dist._neg_logpdf(sample) + corr

        fitted.update(loss, dist)

    return fitted


def fit_to_quartile(dist_names, q1, q2, q3, extra_pros):
    error = np.inf

    for distribution in get_distributions(dist_names):
        if distribution.__class__.__name__ in extra_pros:
            distribution._parametrization(**extra_pros[distribution.__class__.__name__])
            if distribution.__class__.__name__ == "BetaScaled":
                update_bounds_beta_scaled(
                    distribution,
                    extra_pros[distribution.__class__.__name__]["lower"],
                    extra_pros[distribution.__class__.__name__]["upper"],
                )
        if distribution._check_endpoints(q1, q3, raise_error=False):
            none_idx, fixed = get_fixed_params(distribution)

            distribution._fit_moments(
                mean=q2, sigma=(q3 - q1) / 1.35
            )  # pylint:disable=protected-access

            optimize_quartile(distribution, (q1, q2, q3), none_idx, fixed)

            r_error, _ = relative_error(distribution, q1, q3, 0.5)
            if r_error < error:
                fitted_dist = distribution
                error = r_error

    return fitted_dist


def update_bounds_beta_scaled(dist, x_min, x_max):
    dist.lower = x_min
    dist.upper = x_max
    dist.support = (x_min, x_max)
    return dist


def get_penalization(n, dist):
    """
    AIC with a correction for small sample sizes.

    Burnham, K. P.; Anderson, D. R. (2004),
    "Multimodel inference: understanding AIC and BIC in Model Selection"
    shorturl.at/IUWX6
    """
    k = len(dist.params)
    return k + ((k + 1) * k) / (-k + n - 1)


class Loss:
    def __init__(self, size=None):
        self.old_loss = np.inf
        self.dist = None
        self.count = 0
        self.distributions = np.empty(size, dtype=object)
        self.losses = np.empty(size)

    def update(self, loss, dist):
        self.distributions[self.count] = dist
        self.losses[self.count] = loss
        self.count += 1
        if loss < self.old_loss:
            self.old_loss = loss
            self.dist = dist


def get_fixed_params(distribution):
    none_idx = []
    fixed = []
    for idx, p_n in enumerate(distribution.param_names):
        value = getattr(distribution, p_n)
        if value is None:
            none_idx.append(idx)
        else:
            fixed.append(value)
    return none_idx, fixed


def find_kappa(data, mu):
    ere = np.mean(np.cos(mu - data))

    if ere > 0:

        def solve_for_kappa(kappa):
            return i1e(kappa) / i0e(kappa) - ere

        root_res = root_scalar(
            solve_for_kappa, method="brentq", bracket=(np.finfo(float).tiny, 1e16)
        )
        return root_res.root
    else:
        return np.finfo(float).tiny


def find_ppf(dist, q):
    q = np.atleast_1d(q)
    ppf = np.zeros_like(q)
    lower, upper = dist.support
    for idx, q_i in enumerate(q):
        if q_i < 0:
            ppf[idx] = np.nan
        elif q_i > 1:
            ppf[idx] = np.nan
        elif q_i == 0:
            if dist.kind == "discrete":
                ppf[idx] = lower - 1
            else:
                ppf[idx] = lower
        elif q_i == 1:
            ppf[idx] = upper
        else:
            if dist.__class__.__name__ in ["HyperGeometric", "BetaBinomial"]:
                ppf[idx] = _ppf_single(dist, q_i) + 1
            else:
                ppf[idx] = _ppf_single(dist, q_i)
    return ppf[0] if len(ppf) == 1 else ppf


def _ppf_single(dist, q):
    def func(x, dist, q):
        return dist.cdf(x) - q

    factor = 10.0
    left, right = dist.support

    left = min(-factor, right)
    while func(left, dist, q) > 0.0:
        left, right = left * factor, left

    right = max(factor, left)
    while func(right, dist, q) < 0.0:
        left, right = right, right * factor

    return brentq(func, left, right, args=(dist, q))


def get_weighted_rvs(target, size, rng):
    targets = [t[0] for t in target]
    weights = [t[1] for t in target]
    target_rnd_choices = np.random.choice(len(targets), size=size, p=weights)
    samples = [target.rvs(size, random_state=rng) for target in targets]
    return np.choose(target_rnd_choices, samples)



"""Projective predictive elicitation."""

import logging
import numpy as np

from preliz.internal.optimization import optimize_pymc_model
from preliz.ppls.pymc_io import (
    get_model_information,
    get_guess,
    compile_logp,
    backfitting,
    write_pymc_string,
)


_log = logging.getLogger("preliz")


def ppe(model, target, seed=0):
    """
    Projective Predictive Elicitation.

    This is an experimental method under development, use with caution.
    Most likely thing will break.

    Parameters
    ----------
    model : a probabilistic model
        Currently it only works with PyMC model. More PPls coming soon.
    target : a Preliz distribution or list
        Instance of a PreliZ distribution or a list of tuples where each tuple contains a PreliZ
        distribution and a weight.
        This represents the prior predictive distribution **previously** elicited from the user,
        possibly using other Preliz's methods to obtain this distribution, such as maxent,
        roulette, quartile, etc.
        This should represent the domain-knowledge of the user and not any observed dataset.
    seed : int
        A seed to initialize the Random Generator. Default is 0.

    Returns
    -------
    prior : a dictionary
        Prior samples approximating the prior distribution that will induce
        a prior predictive distribution close to ``target``.
    new_priors : a dictionary
        PreliZ Distributions approximating the prior distribution inducing a
        prior predictive distribution close to ``target``.
    pymc_string : a string
        This is the PyMC model string with the new priors.
        Computed by taking the "prior samples" and fit it into the model's prior families
        using MLE.
    """
    _log.info(""""This is an experimental method under development, use with caution.""")

    # Get information from PyMC model
    rng = np.random.default_rng(seed)
    bounds, prior, p_model, var_info, var_info2, draws, free_rvs = get_model_information(model)
    # Initial point for optimization
    guess = get_guess(model, free_rvs)
    # compile PyMC model
    fmodel = compile_logp(model)
    # find prior that induce a prior predictive distribution close to target
    prior = optimize_pymc_model(fmodel, target, draws, prior, guess, bounds, var_info, p_model, rng)
    # Fit the prior into the model's prior
    # So we can write it as a PyMC model
    new_priors = backfitting(prior, p_model, var_info2)

    return prior, new_priors, write_pymc_string(new_priors, var_info2)



import pytest
import numpy as np
from preliz import beta_mode


def test_beta_mode():

    _, dist = beta_mode(0.25, 0.75, 0.5, 0.9)

    assert np.isclose(dist.alpha, 4.94, atol=0.01)
    assert np.isclose(dist.beta, 4.94, atol=0.01)


def test_invalid_mass():
    with pytest.raises(ValueError):
        beta_mode(0.25, 0.75, 0.5, mass=1.1)


def test_invalid_mode():
    with pytest.raises(ValueError):
        beta_mode(0.25, 0.75, 1.5)


def test_invalid_bounds():
    with pytest.raises(ValueError):
        beta_mode(0.75, 0.25, 0.5)


def test_plot_beta_mode():
    _, _ = beta_mode(0.25, 0.75, 0.5, 0.9, plot=True, plot_kwargs={"pointinterval": True})



import importlib
import inspect
import re
from sys import modules

import numpy as np

from preliz import distributions
from .distribution_helper import init_vals


def inspect_source(fmodel):
    source = inspect.getsource(fmodel)
    signature = inspect.signature(fmodel)
    source = re.sub(r"#.*$|^#.*$", "", source, flags=re.MULTILINE)
    default_params = {
        name: (param.default if param.default is not inspect.Parameter.empty else np.nan)
        for name, param in signature.parameters.items()
    }
    model = fmodel(**default_params)
    return source, signature, get_engine(model)


def get_engine(model):
    if getattr(model, "basic_RVs", False):
        return "pymc"
    elif getattr(model, "formula", False):
        return "bambi"
    return "preliz"


def parse_function_for_pred_textboxes(source, signature, engine="preliz"):
    model = {}

    slidify = list(signature.parameters.keys())
    regex = r"\b" + r"\b|\b".join(slidify) + r"\b"

    all_dist_str = dist_as_str()
    matches = match_preliz_dist(all_dist_str, source, engine)

    for match in matches:
        if engine == "bambi":
            dist_name_str = match.group(1)
        else:
            dist_name_str = match.group(2)
        if engine == "bambi":
            arguments = [s.strip() for s in match.group(2).split(",")]
        else:
            arguments = [s.strip() for s in match.group(3).split(",")]
        if engine == "pymc":
            args = pymc_parse_arguments(arguments, regex)
        else:
            args = parse_arguments(arguments, regex)
        for arg in args:
            if arg:
                func, var, idx = arg
                dist = getattr(distributions, dist_name_str)
                model[var] = (dist(**init_vals[dist_name_str]), idx, func)

    return model


def parse_arguments(lst, regex):
    result = []
    for idx, item in enumerate(lst):
        match = re.search(regex, item)
        if match:
            if item.isidentifier():
                result.append((None, match.group(0), idx))
            else:
                if "**" in item:
                    power = item.split("**")[1].strip()
                    result.append((power, match.group(0), idx))
                else:
                    func = item.split("(")[0].split(".")[-1]
                    result.append((func, match.group(0), idx))
    return result


def pymc_parse_arguments(lst, regex):
    result = []
    for idx, item in enumerate(lst):
        match = re.search(regex, item)
        if match:
            if item.isidentifier():
                result.append((None, match.group(0), idx - 1))
            else:
                if "**" in item:
                    power = item.split("**")[1].strip()
                    result.append((power, match.group(0), idx - 1))
                else:
                    func = item.split("(")[0].split(".")[-1]
                    result.append((func, match.group(0), idx - 1))
    return result


def get_prior_pp_samples(fmodel, variables, draws, engine=None, values=None):
    if values is None:
        values = []

    if engine == "preliz":
        obs_rv = variables[-1]  # only one observed for the moment
        pp_samples_ = []
        prior_samples_ = {name: [] for name in variables[:-1]}
        for _ in range(draws):
            for name, value in zip(variables, fmodel(*values)):
                if name == obs_rv:
                    pp_samples_.append(value)
                else:
                    prior_samples_[name].append(value)

        pp_samples = np.stack(pp_samples_)
        prior_samples = {key: np.array(val) for key, val in prior_samples_.items()}
    elif engine == "bambi":
        *prior_samples_, pp_samples = fmodel(*values)
        prior_samples = {name: np.array(val) for name, val in zip(variables[:-1], prior_samples_)}

    return pp_samples, prior_samples


def from_preliz(fmodel):
    source = inspect.getsource(fmodel)
    variables = match_return_variables(source)
    # Find the priors we want to change
    all_dist_str = dist_as_str()
    matches = match_preliz_dist(all_dist_str, source, "preliz")
    # Create a dictionary with the priors
    model = dict_model(matches, variables)

    return variables, model


def from_bambi(fmodel, draws):
    module_name = fmodel.__module__
    module = importlib.import_module(module_name)

    # Get the source code of the original function
    original_source = inspect.getsource(fmodel)

    # Define a pattern to find the line where the model is built
    pattern = re.compile(r"(\s+)([a-zA-Z_]\w*)\s*=\s*.*?Model(.*)")

    # Find the match in the source code
    match = pattern.search(original_source)

    # Extract the indentation and variable name
    indentation = match.group(1)
    variable_name = match.group(2)

    # Find the variables after the return statement
    return_variables = match_return_variables(original_source)

    if return_variables:
        # Build the new source code
        new_source = original_source.replace(
            match.group(0),
            f"{match.group(0)}"
            f"{indentation}{variable_name}.build()\n"
            f"{indentation}variables = [{variable_name}.backend.model.named_vars[v] "
            f"for v in {return_variables}]\n"
            f'{indentation}{", ".join(return_variables)} = pm.draw(variables, draws={draws})',
        )

        # Find the priors we want to change
        all_dist_str = dist_as_str()
        matches = match_preliz_dist(all_dist_str, new_source, "bambi")
        # Create a dictionary with the priors
        model = dict_model(matches, return_variables)

        # Execute the new source code to redefine the function
        exec(new_source, module.__dict__)  # pylint: disable=exec-used
        modified_fmodel = getattr(module, fmodel.__name__)

    return modified_fmodel, return_variables, model


def match_preliz_dist(all_dist_str, source, engine):
    # remove comments
    source = re.sub(r"#.*$|^#.*$", "", source, flags=re.MULTILINE)

    if engine in ["preliz", "pymc"]:
        regex = rf"(.*?({all_dist_str}).*?)\(([^()]*(?:\([^()]*\)[^()]*)*)\)"
    if engine == "bambi":
        regex = rf'\s*(?:\w+\.)?Prior\("({all_dist_str})",\s*((?:\w+=\w+(?:,?\s*)?)*)\s*\)'
    matches = re.finditer(regex, source)
    return matches


def match_return_variables(source):
    match = re.search(r"return (\w+(\s*,\s*\w+)*)", source)
    return [var.strip() for var in match.group(1).split(",")]


def dist_as_str():
    all_distributions = modules["preliz.distributions"].__all__
    return "|".join(all_distributions)


def dict_model(matches, return_variables):
    model = {}
    obs_rv = return_variables[-1]
    for match in matches:
        var_name = match.group(0).split("=")[0].strip()
        if var_name != obs_rv:
            dist = getattr(modules["preliz.distributions"], match.group(2))
            model[var_name] = dist()

    return model



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Discrete
from ..internal.distribution_helper import all_not_none
from ..internal.special import cdf_bounds, ppf_bounds_disc


class DiscreteUniform(Discrete):
    R"""
    Discrete Uniform distribution.

    The pmf of this distribution is

    .. math:: f(x \mid lower, upper) = \frac{1}{upper-lower+1}

    .. plot::
        :context: close-figs


        from preliz import DiscreteUniform, style
        style.use('preliz-doc')
        ls = [1, -2]
        us = [6, 2]
        for l, u in zip(ls, us):
            ax = DiscreteUniform(l, u).plot_pdf()
            ax.set_ylim(0, 0.25)

    ========  ===============================================
    Support   :math:`x \in {lower, lower + 1, \ldots, upper}`
    Mean      :math:`\dfrac{lower + upper}{2}`
    Variance  :math:`\dfrac{(upper - lower + 1)^2 - 1}{12}`
    ========  ===============================================

    Parameters
    ----------
    lower: int
        Lower limit.
    upper: int
        Upper limit (upper > lower).
    """

    def __init__(self, lower=None, upper=None):
        super().__init__()
        self._parametrization(lower, upper)

    def _parametrization(self, lower=None, upper=None):
        self.lower = lower
        self.upper = upper
        self.params = (self.lower, self.upper)
        self.param_names = ("lower", "upper")
        self.params_support = ((-np.inf, np.inf), (-np.inf, np.inf))
        if lower is None:
            self.lower = -np.inf
        if upper is None:
            self.upper = np.inf
        self.support = (self.lower, self.upper)
        if all_not_none(lower, upper):
            self._update(lower, upper)
        else:
            self.lower = lower
            self.upper = upper

    def _get_frozen(self):
        frozen = None
        if all_not_none(self.params):
            frozen = self.dist(self.lower, self.upper + 1)
        return frozen

    def _update(self, lower, upper):
        self.lower = np.floor(lower)
        self.upper = np.ceil(upper)
        self._n = self.upper - self.lower + 1
        self.params = (self.lower, self.upper)
        self.support = (self.lower, self.upper)

        self.is_frozen = True

    def pdf(self, x):
        """
        Compute the probability density function (PDF) at a given point x.
        """
        x = np.asarray(x)
        return nb_pdf(x, self._n, self.lower, self.upper)

    def cdf(self, x):
        """
        Compute the cumulative distribution function (CDF) at a given point x.
        """
        x = np.asarray(x)
        return nb_cdf(x, self._n, self.lower, self.upper)

    def ppf(self, q):
        """
        Compute the percent point function (PPF) at a given probability q.
        """
        q = np.asarray(q)
        return nb_ppf(q, self._n, self.lower, self.upper)

    def logpdf(self, x):
        """
        Compute the log probability density function (log PDF) at a given point x.
        """
        return nb_logpdf(x, self._n, self.lower, self.upper)

    def _neg_logpdf(self, x):
        """
        Compute the neg log_pdf sum for the array x.
        """
        return nb_neg_logpdf(x, self._n, self.lower, self.upper)

    def entropy(self):
        return nb_entropy(self._n)

    def mean(self):
        return (self.upper + self.lower) / 2

    def median(self):
        return np.floor((self.upper + self.lower) / 2)

    def var(self):
        return (self._n**2 - 1) / 12

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 0

    def kurtosis(self):
        return -(6 * (self._n**2 + 1)) / (5 * (self._n**2 - 1))

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.integers(self.lower, self.upper + 1, size)

    def _fit_moments(self, mean, sigma):
        spr = (12 * sigma**2 + 1) ** 0.5
        lower = 0.5 * (2 * mean - spr + 1)
        upper = 0.5 * (2 * mean + spr - 1)
        self._update(lower, upper)

    def _fit_mle(self, sample):
        lower = np.min(sample)
        upper = np.max(sample)
        self._update(lower, upper)


@nb.njit(cache=True)
def nb_cdf(x, n, lower, upper):
    prob = (np.floor(x) - lower + 1) / n
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, n, lower, upper):
    x_vals = np.ceil(q * n + lower - 1)
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_pdf(x, n, lower, upper):
    if lower <= x <= upper:
        return 1 / n
    else:
        return 0


@nb.njit(cache=True)
def nb_entropy(n):
    return np.log(n)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, n, lower, upper):
    if lower <= x <= upper:
        return -np.log(n)
    else:
        return -np.inf


@nb.njit(cache=True)
def nb_neg_logpdf(x, n, lower, upper):
    return -(nb_logpdf(x, n, lower, upper)).sum()
